;**************************** Librera "LCD_4BIT.INC" 
; Estas subrutinas permiten realizar las tareas bsicas de control de un 
;mdulo LCD de 2
; lneas por 16 caracteres, compatible con el modelo LM016L.
; El visualizador LCD est conectado al Puerto B del PIC mediante un bus de 4 
;bits. Las
; conexiones son:
; -    Las 4 lneas superiores del mdulo LCD, pines <DB7B4>  se conectan a 
;las 4
;    lneas superiores del Puerto B del PIC, pines <RB7:RB4>.
; -    Pin RS del LCD a la lnea RA0 del PIC.
; -    Pin R/W del LCD a la lnea RA1 del PIC, o a masa.
; -    Pin Enable del LCD a la lnea RA2 del PIC.
;
; Se utilizan llamadas a subrutinas de retardo de tiempo localizadas en la 
;librera
; RETARDOS.INC.
;
; ZONA DE DATOS 

*********************************************************************

    CBLOCK
    LCD_Dato
    LCD_GuardaDato
    LCD_GuardaTRISB
    LCD_Auxiliar1
    LCD_Auxiliar2
    ENDC

LCD_CaracteresPorLinea    EQU    .16            ; Nmero de caracteres por 

;lnea de la pantalla.

#DEFINE  LCD_PinRS    		PORTA,0
#DEFINE  LCD_PinRW    		PORTA,1
#DEFINE  LCD_PinEnable    	PORTA,2
#DEFINE  LCD_BusDatos    	PORTB

; Subrutina "LCD_Inicializa" 

------------------------------------------------------------
;
; Inicializacin del mdulo LCD: Configura funciones del LCD, produce reset por software,
; borra memoria y enciende pantalla. El fabricante especifica que para garantizar la
; configuracin inicial hay que hacerla como sigue:
;
LCD_Inicializa
    bsf        STATUS,RP0                ; Configura las lneas conectadas al pines RS,
    bcf        LCD_PinRS                ; R/W y E.
    bcf        LCD_PinEnable
    bcf        LCD_PinRW
    bcf        STATUS,RP0
    bcf        LCD_PinRW                ; En caso de que est conectado le indica
                                    ; que se va a escribir en el LCD.
    bcf         LCD_PinEnable            ; Impide funcionamiento del LCD poniendo E=0.
    bcf     	LCD_PinRS                ; Activa el Modo Comando poniendo RS=0.
    call    	Retardo_20ms
    movlw    	b'00110000'   
    call    	LCD_EscribeLCD            ; Escribe el dato en el LCD.
    call    	Retardo_5ms   
    movlw    	b'00110000'   
    call    	LCD_EscribeLCD
    call    	Retardo_200micros
    movlw    	b'00110000'   
    call    	LCD_EscribeLCD
    call    	Retardo_20micros        ; Este retardo es necesario para simular en PROTEUS.
    movlw    	b'00100000'                ; Interface de 4 bits.
    call    	LCD_EscribeLCD
    call    	Retardo_20micros        ; Este retardo es necesario para simular en PROTEUS.
   
; Ahora configura el resto de los parmetros:

    call    LCD_2Lineas4Bits5x7        ; LCD de 2 lneas y caracteres de 5x7 puntos.
    call    LCD_Borra                ; Pantalla encendida y limpia. Cursor al principio
    call    LCD_CursorOFF            ; de la lnea 1. Cursor apagado.
    call    LCD_CursorIncr            ; Cursor en modo incrementar.
    return

; Subrutina "LCD_EscribeLCD" 

-----------------------------------------------------------
;
; Enva el dato del registro de trabajo W al bus de dato y produce un pequeo pulso en el pin
; Enable del LCD. Para no alterar el contenido de las lneas de la parte baja del Puerto B que
; no son utilizadas para el LCD (pines RB3:RB0), primero se lee estas lneas y despus se
; vuelve a enviar este dato sin cambiarlo.

LCD_EscribeLCD
    andlw    b'11110000'                ; Se queda con el nibble alto del dato que es el
    movwf    LCD_Dato                ; que hay que enviar y lo guarda.
    movf     LCD_BusDatos,W            ; Lee la informacin actual de la parte baja
    andlw    b'00001111'                ; del Puerto B, que no se debe alterar.
    iorwf    LCD_Dato,F                ; Enviar la parte alta del dato de entrada
                                    ; y en la parte baja lo que haba antes.
    bsf      STATUS,RP0                ; Acceso al Banco 1.
    movf     TRISB,W                    ; Guarda la configuracin que tena antes TRISB.
    movwf    LCD_GuardaTRISB
    movlw    b'00001111'                ; Las 4 lneas inferiores del Puerto B se dejan
    andwf    PORTB,F                    ; como estaban y las 4 superiores como salida.
    bcf      STATUS,RP0                ; Acceso al Banco 0.
    movf     LCD_Dato,W                ; Recupera el dato a enviar.
    movwf    LCD_BusDatos            ; Enva el dato al mdulo LCD.
    bsf      LCD_PinEnable            ; Permite funcionamiento del LCD mediante un pequeo
    bcf      LCD_PinEnable            ; pulso y termina impidiendo el funcionamiento del LCD.
    bsf      STATUS,RP0                ; Acceso al Banco 1. Restaura el antiguo valor en
    movf     LCD_GuardaTRISB,W        ; la configuracin del Puerto B.
    movwf    TRISB
    bcf      STATUS,RP0                ; Acceso al Banco 0.
    return

; Subrutinas variadas para el control del mdulo LCD 

-----------------------------------------
;
;Los comandos que pueden ser ejecutados son:
;
LCD_CursorIncr                        ; Cursor en modo incrementar.
    movlw    b'00000110'
    goto    LCD_EnviaComando
LCD_Linea1                            ; Cursor al principio de la Lnea 1.
    movlw    b'10000000'                ; Direccin 00h de la DDRAM
    goto    LCD_EnviaComando
LCD_Linea2                            ; Cursor al principio de la Lnea 2.
    movlw    b'11000000'                ; Direccin 40h de la DDRAM
    goto    LCD_EnviaComando
LCD_Linea3                            ; Cursor al principio de la Lnea 3
    movlw    b'10010100'                ; Direccin 14h de la DDRAM
    goto    LCD_EnviaComando
LCD_Linea4                            ; Cursor al principio de la Lnea 4
    movlw    b'11010100'                ; Direccin 54h de la DDRAM
    goto    LCD_EnviaComando
LCD_PosicionLinea1                    ; Cursor a posicin de la Lnea 1, a partir de la
    iorlw    b'10000000'                ; direccin 00h de la DDRAM ms el valor del
    goto    LCD_EnviaComando        ; registro W.
LCD_PosicionLinea2                    ; Cursor a posicin de la Lnea 2, a partir de la
    iorlw    b'11000000'                ; direccin 40h de la DDRAM ms el valor del
    goto    LCD_EnviaComando        ; registro W.
LCD_OFF                                ; Pantalla apagada.
    movlw    b'00001000'
    goto    LCD_EnviaComando
LCD_CursorON                        ; Pantalla encendida y cursor encendido.
    movlw    b'00001110'
    goto    LCD_EnviaComando
LCD_CursorOFF                        ; Pantalla encendida y cursor apagado.
    movlw    b'00001100'
    goto    LCD_EnviaComando
LCD_Borra                            ; Borra toda la pantalla, memoria DDRAM y pone el
    movlw    b'00000001'                ; cursor a principio de la lnea 1.
    goto    LCD_EnviaComando
LCD_2Lineas4Bits5x7                    ; Define la pantalla de 2 lneas, con caracteres
    movlw    b'00101000'                ; de 5x7 puntos y conexin al PIC mediante bus de
;    goto    LCD_EnviaComando        ; 4 bits.

; Subrutinas "LCD_EnviaComando" y "LCD_Caracter" 

------------------------------------
;
; "LCD_EnviaComando". Escribe un comando en el registro del mdulo LCD. La palabra de
; comando ha sido entregada a travs del registro W.  Trabaja en Modo Comando.
; "LCD_Caracter". Escribe en la memoria DDRAM del LCD el carcter ASCII introducido a
; a travs del registro W. Trabaja en Modo Dato.
;
LCD_EnviaComando
    bcf     LCD_PinRS                ; Activa el Modo Comando, poniendo RS=0.
    goto    LCD_Envia
LCD_Caracter
    bsf     LCD_PinRS                ; Activa el "Modo Dato", poniendo RS=1.
    call    LCD_CodigoCGROM            ; Obtiene el cdigo para correcta visualizacin.
LCD_Envia
    movwf   LCD_GuardaDato            ; Guarda el dato a enviar.
    call    LCD_EscribeLCD            ; Primero enva el nibble alto.
    swapf   LCD_GuardaDato,W        ; Ahora enva el nibble bajo. Para ello pasa el
                                    ; nibble bajo del dato a enviar a parte alta del byte.
    call    LCD_EscribeLCD            ; Se enva al visualizador LCD.
    btfss   LCD_PinRS                ; Debe garantizar una correcta escritura manteniendo
    call    Retardo_2ms                ; 2 ms en modo comando y 50 s en modo cracter.
    call    Retardo_50micros
    return   

; Subrutina "LCD_CodigoCGROM" 

-----------------------------------------------------------
;
; A partir del carcter ASCII nmero 127 los cdigos de los caracteres definidos en la
; tabla CGROM del LM016L no coinciden con los cdigos ASCII. As por ejemplo, el cdigo
; ASCII de la "" en la tabla CGRAM del LM016L es EEh.
;
; Esta subrutina convierte los cdigos ASCII de la "", "" y otros, a cdigos CGROM para que
; que puedan ser visualizado en el mdulo LM016L.
;
; Entrada:    En (W) el cdigo ASCII del carcter que se desea visualizar.
; Salida:    En (W) el cdigo definido en la tabla CGROM.

LCD_CodigoCGROM
    movwf    LCD_Dato                ; Guarda el valor del carcter y comprueba si es
LCD_EnheMinuscula                    ; un carcter especial.
    sublw    ''                     ; Es la ""?
    btfss    STATUS,Z
    goto    LCD_EnheMayuscula        ; No es "".
    movlw    b'11101110'                ; Cdigo CGROM de la "".
    movwf    LCD_Dato
    goto    LCD_FinCGROM
LCD_EnheMayuscula
    movf    LCD_Dato,W                ; Recupera el cdigo ASCII de entrada.
    sublw    ''                     ; Es la ""?
    btfss    STATUS,Z
    goto    LCD_Grado                ; No es "".
    movlw    b'11101110'                ; Cdigo CGROM de la "". (No hay smbolo para
    movwf    LCD_Dato                ; la "" mayscula en la CGROM).
    goto    LCD_FinCGROM   
LCD_Grado
    movf    LCD_Dato,W                ; Recupera el cdigo ASCII de entrada.
    sublw    ''                     ; Es el smbolo ""?
    btfss    STATUS,Z
    goto    LCD_FinCGROM            ; No es "".
    movlw    b'11011111'                ; Cdigo CGROM del smbolo "".
    movwf    LCD_Dato
LCD_FinCGROM
    movf    LCD_Dato,W                ; En (W) el cdigo buscado.
    return

; Subrutina "LCD_DosEspaciosBlancos" y "LCD_LineaBlanco" 

--------------------------------
;
; Visualiza espacios en blanco.

LCD_LineaEnBlanco
    movlw    LCD_CaracteresPorLinea
    goto    LCD_EnviaBlancos
LCD_UnEspacioBlanco
    movlw    .1
    goto    LCD_EnviaBlancos
LCD_DosEspaciosBlancos
    movlw    .2
    goto    LCD_EnviaBlancos
LCD_TresEspaciosBlancos
    movlw    .3
LCD_EnviaBlancos
    movwf    LCD_Auxiliar1            ; (LCD_Auxiliar1) se utiliza como contador.
LCD_EnviaOtroBlanco   
    movlw    ' '                        ; Esto es un espacio en blanco.
    call    LCD_Caracter            ; Visualiza tanto espacios en blanco como se
    decfsz    LCD_Auxiliar1,F            ; haya cargado en (LCD_Auxiliar1).
    goto    LCD_EnviaOtroBlanco
    return

; Subrutinas "LCD_ByteCompleto" y "LCD_Byte" 

--------------------------------------------
;
; Subrutina "LCD_ByteCompleto", visualiza el byte que almacena el registro W en el
; lugar actual de la pantalla. Por ejemplo, si (W)=b'10101110' visualiza "AE".
; Subrutina "LCD_Byte" igual que la anterior, pero en caso de que el nibble alto sea cero
; visualiza en su lugar un espacio en blanco. Por ejemplo si (W)=b'10101110' visualiza "AE"
; y si (W)=b'00001110', visualiza " E" (un espacio blanco delante).
;
; Utilizan la subrutina "LCD_Nibble" que se analiza ms adelante.
;
LCD_Byte
    movwf    LCD_Auxiliar2            ; Guarda el valor de entrada.
    andlw    b'11110000'                ; Analiza si el nibble alto es cero.
    btfss    STATUS,Z                ; Si es cero lo apaga.
    goto    LCD_VisualizaAlto        ; No es cero y lo visualiza.
    movlw    ' '                        ; Visualiza un espacio en blanco.
    call    LCD_Caracter
    goto    LCD_VisualizaBajo

LCD_ByteCompleto
    movwf    LCD_Auxiliar2            ; Guarda el valor de entrada.
LCD_VisualizaAlto
    swapf    LCD_Auxiliar2,W            ; Pone el nibble alto en la parte baja.
    call    LCD_Nibble                ; Lo visualiza.
LCD_VisualizaBajo
    movf    LCD_Auxiliar2,W            ; Repite el proceso con el nibble bajo.
;    call    LCD_Nibble                ; Lo visualiza.
;    return

; Subrutina "LCD_Nibble" 

----------------------------------------------------------------
;
; Visualiza en el lugar actual de la pantalla, el valor hexadecimal que almacena en el nibble
; bajo del registro W. El nibble alto de W no es tenido en cuenta. Ejemplos:
; - Si (W)=b'01010110', se visualizar "6".
; - Si (W)=b'10101110', se visualizar "E".
;
LCD_Nibble
    andlw    b'00001111'                ; Se queda con la parte baja.
    movwf    LCD_Auxiliar1            ; Lo guarda.
    sublw    0x09                    ; Comprueba si hay que representarlo con letra.
    btfss    STATUS,C   
    goto    LCD_EnviaByteLetra
    movf    LCD_Auxiliar1,W
    addlw    '0'                        ; El nmero se pasa a carcter ASCII sumndole
    goto     LCD_FinVisualizaDigito    ; el ASCII del cero y lo visualiza.
LCD_EnviaByteLetra
    movf    LCD_Auxiliar1,W
    addlw    'A'-0x0A                ; S, por tanto, se le suma el ASCII de la 'A'.
LCD_FinVisualizaDigito
    goto    LCD_Caracter            ; Y visualiza el carcter. Se hace con un "goto"
                                    ; para no sobrecargar la pila.
